/*
 * libsyncml - A syncml protocol implementation
 * Copyright (C) 2005  Armin Bauer <armin.bauer@opensync.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; version 
 * 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 */

#include "../support.h"

#define NUM_SESSIONS 30

/* ************************************
 * *********** OBEX SERVER ************
 * ************************************
 *         (mobile simulator)
 */

#include <libsyncml/sml_transport_internals.h>

SmlTransport *mobile_tsp = NULL;

void obex_mobile_send(SmlLink *link, const char *data, unsigned int size)
{
	SmlError *error = NULL;

	SmlTransportData *tsp_data = smlTransportDataNew(g_strdup(data), size, SML_MIMETYPE_XML, FALSE, &error);
	fail_unless(tsp_data != NULL, NULL);
	fail_unless(error == NULL, NULL);

	fail_unless(smlTransportSend(mobile_tsp, link, tsp_data, &error), smlErrorPrint(&error));
	fail_unless(error == NULL, NULL);
	
	smlTransportDataDeref(tsp_data);
}

SmlBool equal_messages(const char *now, unsigned int size, const char *expected)
{
	smlTrace(TRACE_ENTRY, "%s", __func__);
	GRegex *regex = NULL;
	GError *error = NULL;
	char *now_1;
	char *now_2;
	char *org_1;
	char *org_2;

	/* ignore SessionID */

	regex = g_regex_new("<SessionID>[0-9]+</SessionID>", 0, 0, &error);
	fail_unless(regex != NULL, "SessionID filter cannot be established.");

	now_1 = g_regex_replace_literal(regex, now, size,
			0, "<SessionID>2346</SessionID>", 0,
			&error);
	fail_unless(now_1 != NULL, "Cannot filter SessionID from new message.");

	org_1 = g_regex_replace_literal(regex, expected, strlen(expected),
			0, "<SessionID>2346</SessionID>", 0,
			&error);
	fail_unless(org_1 != NULL, "Cannot filter SessionID from original message.");

	g_regex_unref(regex);

	/* ignore trailing whitespaces */

	regex = g_regex_new("[\\s\\n]*$", G_REGEX_DOLLAR_ENDONLY, 0, &error);
	fail_unless(regex != NULL, "Trailing whitespace filter cannot be established.");

	now_2 = g_regex_replace_literal(regex, now_1, strlen(now_1), 0, "", 0, &error);
	fail_unless(now_2 != NULL, "Cannot filter trailing whitespaces from new message.");
	smlSafeCFree(&now_1);

	org_2 = g_regex_replace_literal(regex, org_1, strlen(org_1), 0, "", 0, &error);
	fail_unless(org_2 != NULL, "Cannot filter trailing whitespaces from original message.");
	smlSafeCFree(&org_1);

	g_regex_unref(regex);

	/* ignore newlines (not carriage return) */

	regex = g_regex_new(">\\n<", 0, 0, &error);
	fail_unless(regex != NULL, "Useless newline filter cannot be established.");

	now_1 = g_regex_replace_literal(regex, now_2, strlen(now_2), 0, "><", 0, &error);
	fail_unless(now_1 != NULL, "Cannot filter useless newlines from new message.");
	smlSafeCFree(&now_2);

	org_1 = g_regex_replace_literal(regex, org_2, strlen(org_2), 0, "><", 0, &error);
	fail_unless(org_1 != NULL, "Cannot filter useless newlines from original message.");
	smlSafeCFree(&org_2);

	g_regex_unref(regex);

	/* ignore numbered next values */

	regex = g_regex_new("<Next xmlns=\"syncml:metinf\">[0-9]+</Next>", 0, 0, &error);
	fail_unless(regex != NULL, "Numbered Next filter cannot be established.");

	now_2 = g_regex_replace_literal(regex, now_1, strlen(now_1),
			0, "<Next xmlns=\"syncml:metinf\">1357</Next>", 0, &error);
	fail_unless(now_2 != NULL, "Cannot filter numbered Next from new message.");
	smlSafeCFree(&now_1);

	org_2 = g_regex_replace_literal(regex, org_1, strlen(org_1),
			0, "<Next xmlns=\"syncml:metinf\">1357</Next>", 0, &error);
	fail_unless(org_2 != NULL, "Cannot filter numbered Next from original message.");
	smlSafeCFree(&org_1);

	g_regex_unref(regex);

	/* ignore timestamp next values */

	regex = g_regex_new("<Next xmlns=\"syncml:metinf\">[0-9]{8}T[0-9]{6}Z</Next>", 0, 0, &error);
	fail_unless(regex != NULL, "Timestamp Next filter cannot be established.");

	now_1 = g_regex_replace_literal(regex, now_2, strlen(now_2),
			0, "<Next xmlns=\"syncml:metinf\">20080101T140000Z</Next>", 0, &error);
	fail_unless(now_1 != NULL, "Cannot filter timestamp Next from new message.");
	smlSafeCFree(&now_2);

	org_1 = g_regex_replace_literal(regex, org_2, strlen(org_2),
			0, "<Next xmlns=\"syncml:metinf\">20080101T140000Z</Next>", 0, &error);
	fail_unless(org_1 != NULL, "Cannot filter timestamp Next from original message.");
	smlSafeCFree(&org_2);

	g_regex_unref(regex);

	/* compare strings */

	SmlBool notEqual = strcmp(now_1, org_1);

	if (notEqual)
		smlTrace(TRACE_INTERNAL, "%s: FAILED\n==>%s<==\n-----\n==>%s<==", __func__, now_1, org_1);
	smlSafeCFree(&now_1);
	smlSafeCFree(&org_1);

	smlTrace(TRACE_EXIT, "%s - %d", __func__, (!notEqual));
	return (!notEqual);
}

SmlBool _recv_server_event(SmlTransport *tsp, SmlLink *link, SmlTransportEventType type, SmlTransportData *data, SmlError *error, void *userdata)
{
	smlTrace(TRACE_ENTRY, "%s()", __func__);
		
	switch (type) {
		case SML_TRANSPORT_EVENT_DATA:

			if (mobile_state == TEST_MOBILE_STATE_WAIT_FOR_NOTIFICATION) {
				smlTrace(TRACE_INTERNAL, "%s: mobile received notification", __func__);
				if (!equal_messages(data->data, data->size, libsyncml_msg_notification))
					fail("wrong notification message");
				mobile_state = TEST_MOBILE_STATE_WAIT_FOR_ALERT;
				obex_mobile_send(link, mobile_msg_alert, strlen(mobile_msg_alert));
			} else if (mobile_state == TEST_MOBILE_STATE_WAIT_FOR_ALERT) {
				smlTrace(TRACE_INTERNAL, "%s: mobile received alert", __func__);
				if (!equal_messages(data->data, data->size, libsyncml_msg_alert))
					fail("wrong alert message");
				mobile_state = TEST_MOBILE_STATE_WAIT_FOR_SYNC;
				obex_mobile_send(link, mobile_msg_sync, strlen(mobile_msg_sync));
			} else if (mobile_state == TEST_MOBILE_STATE_WAIT_FOR_SYNC) {
				smlTrace(TRACE_INTERNAL, "%s: mobile received sync", __func__);
				if (!equal_messages(data->data, data->size, libsyncml_msg_sync))
					fail("wrong sync message");
				mobile_state = TEST_MOBILE_STATE_WAIT_FOR_END;
				obex_mobile_send(link, mobile_msg_map, strlen(mobile_msg_map));
			} else if (mobile_state == TEST_MOBILE_STATE_WAIT_FOR_END) {
				smlTrace(TRACE_INTERNAL, "%s: mobile received end", __func__);
				if (!equal_messages(data->data, data->size, libsyncml_msg_end))
					fail("wrong end message");
				mobile_state = TEST_MOBILE_STATE_FINISHED;
				smlTransportDisconnect(mobile_tsp, link, &error);
			} else {
				fail("Unexpected mobile state.");
			}
			break;
		case SML_TRANSPORT_EVENT_CONNECT_DONE:
			mobile_state = TEST_MOBILE_STATE_CONNECTED;
			fail_unless(libsyncml_msg_notification != NULL, "Missing notification message");
			mobile_state = TEST_MOBILE_STATE_WAIT_FOR_NOTIFICATION;
			break;
		case SML_TRANSPORT_EVENT_DISCONNECT_DONE:
			mobile_state = TEST_MOBILE_STATE_DISCONNECTED;
			break;
		default:
			fail(NULL);
	}
	
	smlTrace(TRACE_EXIT, "%s()", __func__);
	return TRUE;
}

char *obex_port;
void obex_mobile_connect()
{
	SmlError *error = NULL;
	mobile_tsp = smlTransportNew(SML_TRANSPORT_OBEX_SERVER, &error);

	obex_port = g_strdup_printf("%d", g_random_int_range(10001, 19999));
	
	fail_unless(smlTransportSetConfigOption(mobile_tsp, "PORT", obex_port, &error), NULL);
	fail_unless(smlTransportSetConnectionType(mobile_tsp, SML_TRANSPORT_CONNECTION_TYPE_NET, &error), NULL);

	fail_unless(smlTransportInitialize(mobile_tsp, &error), NULL);

	smlTransportSetEventCallback(mobile_tsp, _recv_server_event, GINT_TO_POINTER(1));
}

void obex_mobile_disconnect()
{
	SmlError *error = NULL;
	while (mobile_state != TEST_MOBILE_STATE_DISCONNECTED)
	{
		usleep(100);
	}
	smlTransportStop(mobile_tsp);
	fail_unless(smlTransportFinalize(mobile_tsp, &error), NULL);
	smlTransportFree(mobile_tsp);
}

